/* Time-stamp: <2006-02-04 21:00:33 jcs>
|
|  Copyright (C) 2002-2005 Jorg Schuler <jcsjcs at users sourceforge net>
|  Part of the gtkpod project.
|
|  URL: http://www.gtkpod.org/
|  URL: http://gtkpod.sourceforge.net/
|
|  This program is free software; you can redistribute it and/or modify
|  it under the terms of the GNU General Public License as published by
|  the Free Software Foundation; either version 2 of the License, or
|  (at your option) any later version.
|
|  This program is distributed in the hope that it will be useful,
|  but WITHOUT ANY WARRANTY; without even the implied warranty of
|  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|  GNU General Public License for more details.
|
|  You should have received a copy of the GNU General Public License
|  along with this program; if not, write to the Free Software
|  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
|
|  iTunes and iPod are trademarks of Apple
|
|  This product is not supported/written/published by Apple!
|
|  $Id: prefs.c,v 1.157 2006/02/04 12:04:52 jcsjcs Exp $
*/

/* -------------------------------------------------------------------
 *
 * HOWTO add a new_option to the prefs dialog
 *
 * - add the desired option to the prefs window using glade-2
 *
 * - modify the cfg structure in prefs.h accordingly
 *
 * - set the default value of new_option in cfg_new() in prefs.c
 *
 * - add function prefs_get_new_option() and
 *   prefs_set_new_option() to prefs.[ch]. These functions are
 *   called from within gtkpod to query/set the state of the option.
 *   prefs_set_new_option() should verify that the value passed is
 *   valid.
 *
 * - add a callback on_new_option_*() to prefs_windows.c to set the
 *   new value in the temporary prefs struct.
 *   The value is applied to the actual prefs when pressing the "OK"
 *   or "Apply" button in the prefs window.
 *
 * - if your option is a pointer to data, make sure the data is copied
 *   in clone_prefs() in prefs.c
 *
 * - add code to prefs_window_create() in prefs_window.c to set the
 *   correct state of the option in the prefs window.
 *
 * - add code to prefs_window_set() in prefs_window.c to actually take
 *   over the new state when closing the prefs window
 *
 * - add code to write_prefs_to_file_desc() to write the state of
 *   new_option to the prefs file
 *
 * - add code to read_prefs_from_file_desc() to read the
 *   new_option from the prefs file
 *
 * - if you want new_option to be a command line option as well, add
 *   code to usage() and read_prefs()
 *
 * ---------------------------------------------------------------- */

/* FIXME: simplify code to make adding of new options easier:
		  prefs_window_create()
		  write_prefs_to_file_desc()
		  read_prefs_from_file_desc()
*/

/* This tells Alpha OSF/1 not to define a getopt prototype in <stdio.h>.
   Ditto for AIX 3.2 and <stdlib.h>.  */
#ifndef _NO_PROTO
# define _NO_PROTO
#endif

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_GETOPT_LONG_ONLY
#  include <getopt.h>
#else
#  include "getopt.h"
#endif

#include "clientserver.h"
#include "display.h"
#include "info.h"
#include "misc.h"
#include "misc_track.h"
#include "prefs.h"
#include "podcast.h"

/* static function declarations */
static void prefs_write_hash_values (FILE *fp);

/* config struct */
static struct cfg *cfg = NULL;

/* enum for reading of options */
enum {
  GP_HELP,
  GP_PLAYCOUNT,
  GP_MOUNT,
  GP_AUTO,
  GP_OFFLINE
};


/* need to convert to locale charset before printing to console */
#define usage_fpf(file, ...) do { gchar *utf8=g_strdup_printf (__VA_ARGS__); gchar *loc=g_locale_from_utf8 (utf8, -1, NULL, NULL, NULL); fprintf (file, "%s", loc); g_free (loc); g_free (utf8);} while (FALSE)

static void usage (FILE *file)
{
  usage_fpf(file, _("gtkpod version %s usage:\n"), VERSION);
  usage_fpf(file, _("  -h, --help:   display this message\n"));
  usage_fpf(file, _("  -p <filename>:increment playcount for file by one\n"));
  usage_fpf(file, _("  -m path:      define the mountpoint of your iPod\n"));
  usage_fpf(file, _("  --mountpoint: same as '-m'.\n"));
  usage_fpf(file, _("  -a:           import database automatically after start.\n"));
  usage_fpf(file, _("  --auto:       same as '-a'.\n"));
  usage_fpf(file, _("  -o:           use offline mode. No changes are exported to the iPod,\n                but to ~/.gtkpod/ instead. iPod is updated if 'Sync' is\n                used with 'Offline' deactivated.\n"));
  usage_fpf(file, _("  --offline:    same as '-o'.\n"));
}


struct cfg *cfg_new(void)
{
    struct cfg *mycfg = NULL;
    gchar curdir[PATH_MAX], *str;
    gchar *cfgdir;
    gint i;

    cfgdir = prefs_get_cfgdir ();

    mycfg = g_malloc0 (sizeof (struct cfg));
    if(getcwd(curdir, PATH_MAX))
    {
	mycfg->last_dir.browse = g_strdup (curdir);
    }
    else
    {
	mycfg->last_dir.browse = convert_filename ("~/");
    }
    if((str = getenv("IPOD_MOUNTPOINT")))
    {
	mycfg->ipod_mount = convert_filename (str);
    }
    else
    {
	mycfg->ipod_mount = g_strdup("/mnt/ipod");
    }

    mycfg->charset = NULL;
    mycfg->deletion.track = TRUE;
    mycfg->deletion.ipod_file = TRUE;
    mycfg->deletion.local_file = TRUE;
    mycfg->deletion.database = TRUE;
    mycfg->deletion.syncing = TRUE;
    mycfg->md5tracks = TRUE;
    mycfg->update_existing = FALSE;
    mycfg->block_display = FALSE;
    mycfg->autoimport = FALSE;
    for (i=0; i<SORT_TAB_MAX; ++i)
    {
	mycfg->st[i].autoselect = TRUE;
	mycfg->st[i].category = (i<ST_CAT_NUM ? i:0);
	mycfg->st[i].sp_or = FALSE;
	mycfg->st[i].sp_rating = FALSE;
	mycfg->st[i].sp_rating_state = 0;
	mycfg->st[i].sp_playcount = FALSE;
	mycfg->st[i].sp_playcount_low = 0;
	mycfg->st[i].sp_playcount_high = -1;
	mycfg->st[i].sp_played = FALSE;
	mycfg->st[i].sp_played_state = g_strdup (">4w");
	mycfg->st[i].sp_modified = FALSE;
	mycfg->st[i].sp_modified_state = g_strdup ("<1d");
	mycfg->st[i].sp_added = FALSE;
	mycfg->st[i].sp_added_state = g_strdup ("<1d");
	mycfg->st[i].sp_autodisplay = FALSE;
    }
    mycfg->mpl_autoselect = TRUE;
    mycfg->offline = FALSE;
    mycfg->write_extended_info = TRUE;
    mycfg->id3_write = FALSE;
    mycfg->id3_write_id3v24 = FALSE;
    mycfg->size_gtkpod.x = 600;
    mycfg->size_gtkpod.y = 500;
    mycfg->size_cal.x = 500;
    mycfg->size_cal.y = 350;
    mycfg->size_conf_sw.x = 300;
    mycfg->size_conf_sw.y = 300;
    mycfg->size_conf.x = 300;
    mycfg->size_conf.y = -1;
    mycfg->size_dirbr.x = 300;
    mycfg->size_dirbr.y = 400;
    mycfg->size_prefs.x = -1;
    mycfg->size_prefs.y = 480;
    mycfg->size_info.x = 510;
    mycfg->size_info.y = 300;
    for (i=0; i<TM_NUM_COLUMNS; ++i)
    {
	mycfg->tm_col_width[i] = 80;
	mycfg->col_visible[i] = FALSE;
	mycfg->col_order[i] = i;
    }
    mycfg->col_visible[TM_COLUMN_ARTIST] = TRUE;
    mycfg->col_visible[TM_COLUMN_ALBUM] = TRUE;
    mycfg->col_visible[TM_COLUMN_TITLE] = TRUE;
    mycfg->col_visible[TM_COLUMN_GENRE] = TRUE;
    mycfg->col_visible[TM_COLUMN_PLAYCOUNT] = TRUE;
    mycfg->col_visible[TM_COLUMN_RATING] = TRUE;
    for (i=0; i<TM_NUM_TAGS_PREFS; ++i)
    {
	mycfg->autosettags[i] = FALSE;
    }
    mycfg->autosettags[TM_COLUMN_TITLE] = TRUE;
    mycfg->readtags = TRUE;
    mycfg->parsetags = FALSE;
    mycfg->parsetags_overwrite = FALSE;
    mycfg->parsetags_template = g_strdup ("%a - %A/%T %t.mp3;%t.wav");
    mycfg->coverart = TRUE;
    mycfg->coverart_template = g_strdup ("%A;folder.jpg");
    for (i=0; i<PANED_NUM; ++i)
    {
	mycfg->paned_pos[i] = -1;  /* -1 means: let gtk worry about position */
    }
    mycfg->show_duplicates = TRUE;
    mycfg->show_updated = TRUE;
    mycfg->show_non_updated = TRUE;
    mycfg->show_sync_dirs = TRUE;
    mycfg->sync_remove = TRUE;
    mycfg->display_toolbar = TRUE;
    mycfg->toolbar_style = GTK_TOOLBAR_BOTH;
    mycfg->display_tooltips_main = TRUE;
    mycfg->display_tooltips_prefs = TRUE;
    mycfg->update_charset = FALSE;
    mycfg->write_charset = FALSE;
    mycfg->add_recursively = TRUE;
    mycfg->sort_tab_num = 2;
    mycfg->last_prefs_page = 0;
    mycfg->statusbar_timeout = STATUSBAR_TIMEOUT;
    for (i=0; i<PATH_NUM; ++i)
    {  /* the switch() will cause the compiler to complain if we
	  forget to initialize one of the paths */
	switch ((PathType)i)
	{
	case PATH_PLAY_NOW:
	    mycfg->path[i] = g_strdup ("xmms %s"); break;
	case PATH_PLAY_ENQUEUE:
	    mycfg->path[i] = g_strdup ("xmms -e %s"); break;
	case PATH_MP3GAIN:
	case PATH_SYNC_CONTACTS:
	case PATH_SYNC_CALENDAR:
	case PATH_SYNC_NOTES:
	case PATH_MSERV_MUSIC_ROOT:
	    mycfg->path[i] = g_strdup (""); break;
	case PATH_MSERV_TRACKINFO_ROOT:
	    mycfg->path[i] = g_strdup ("/var/lib/mserv/trackinfo/");
	case PATH_NUM:
	    break;
	}
    }
    mycfg->unused_gboolean3 = FALSE;
    mycfg->concal_autosync = FALSE;
    mycfg->tmp_disable_sort = TRUE;
    mycfg->startup_messages = TRUE;
    mycfg->automount = FALSE;
    mycfg->info_window = FALSE;
    mycfg->multi_edit = FALSE;
    mycfg->multi_edit_title = TRUE;
    mycfg->not_played_track = TRUE;
    mycfg->misc_track_nr = 25;
    mycfg->sortcfg.pm_sort = SORT_NONE;
    mycfg->sortcfg.st_sort = SORT_NONE;
    mycfg->sortcfg.tm_sort = SORT_NONE;
    mycfg->sortcfg.tm_sortcol = TM_COLUMN_TITLE;
    mycfg->sortcfg.pm_autostore = FALSE;
    mycfg->sortcfg.tm_autostore = FALSE;
    mycfg->sortcfg.case_sensitive = FALSE;
    mycfg->mserv_use = FALSE;
    mycfg->mserv_report_probs = TRUE;
    mycfg->mserv_username = g_strdup ("");

    mycfg->pc_dir = g_build_filename (cfgdir, "podcasts", NULL);
    mycfg->pc_del_age = FALSE;
    mycfg->pc_del_age_val = 1;
    mycfg->pc_del_copied = FALSE;
    mycfg->pc_auto_fetch = FALSE;
    mycfg->pc_log = TRUE;
    mycfg->pc_log_file = g_build_filename (cfgdir, "podcast.log", NULL);
    mycfg->pc_auto_sync = FALSE;
    mycfg->pc_ipod_del_age = FALSE;
    mycfg->pc_ipod_del_age_val = 1;
    mycfg->pc_ipod_del_played = FALSE;
    mycfg->pc_ipod_inc_date = FALSE;
    mycfg->pc_change_genre = FALSE;

    g_free (cfgdir);

    return(mycfg);
}


/* Compare strlen(@arg) chars of @arg with @line. Return strlen (@arg)
   in @off if @off != NULL */
static gint arg_comp (const gchar *line, const gchar *arg, gint *off)
{
    if (arg && line)
    {
	gint len = strlen (arg);
	if (off) *off = len;
	return g_ascii_strncasecmp (line, arg, len);
    }
    else
    {
	if (*off)   *off = 0;
	return 0;
    }
}


/* sort order was reversed between V0.82_CVS and V0.83.CVS */
static gint correct_sort (gint sort)
{
    if (cfg && (cfg->version < 0.83))
    {
	switch (sort)
	{
	case SORT_ASCENDING:
	    sort = SORT_DESCENDING;
	    break;
	case SORT_DESCENDING:
	    sort = SORT_ASCENDING;
	    break;
	}
    }
    return sort;
}

/* default ignore strings -- must end with a space */
static char* sort_ign_strings[] =
{
    "the ",
    "a ",
/*    "le ", 
    "la ", 
    "les ", 
    "lo ", 
    "los ",
    "der ",
    "die ",
    "das ",*/ /* will make sorting very slow -- only add the words you
                 really want to skip */
    SORT_IGNORE_STRINGS_END,  /* end marker */
    NULL,
};


static void
read_prefs_from_file_desc(FILE *fp)
{
    gchar buf[PATH_MAX];
    gchar *line, *arg, *bufp;
    gint len, off, i;

    /* set ignore strings */
    for (i=0; sort_ign_strings[i]; ++i)
    {
	bufp = g_strdup_printf ("sort_ign_string_%d", i);
	prefs_set_string_value (bufp, sort_ign_strings[i]);
	g_free (bufp);
    }
    /* set ignore fields (ignore above words for artist) */
    bufp = g_strdup_printf ("sort_ign_field_%d", T_ARTIST);
    prefs_set_int_value (bufp, 1);
    g_free (bufp);

    if(fp)
    {
      while (fgets (buf, PATH_MAX, fp))
	{
	  /* allow comments */
	  if ((buf[0] == ';') || (buf[0] == '#')) continue;
	  arg = strchr (buf, '=');
	  if (!arg || (arg == buf))
	    {
	      gtkpod_warning (_("Error while reading prefs: %s\n"), buf);
	      continue;
	    }
	  /* skip whitespace */
	  bufp = buf;
	  while (g_ascii_isspace(*bufp)) ++bufp;
	  line = g_strndup (buf, arg-bufp);
	  ++arg;
	  len = strlen (arg); /* remove newline */
	  if((len>0) && (arg[len-1] == 0x0a))  arg[len-1] = 0;
	  /* skip whitespace */
	  while (g_ascii_isspace(*arg)) ++arg;
	  if(g_ascii_strcasecmp (line, "version") == 0)
	  {
	      cfg->version = g_ascii_strtod (arg, NULL);
	  }
	  else if(g_ascii_strcasecmp (line, "mountpoint") == 0)
	  {
	      prefs_set_ipod_mount (arg);
	  }
	  else if((arg_comp (line, "toolpath", &off) == 0) ||
		  (arg_comp (line, "path", &off) == 0))
	  {
	      gint i = atoi (line+off);
	      prefs_set_path (i, arg);
	      if ((i == PATH_PLAY_NOW) && (cfg->version < 0.87))
	      {  /* default changed from "xmms -p %s" to "xmms %s" which
		    avoids xmms from hanging -- thanks to Chris	Vine */
		  if (strcmp (arg, "xmms -p %s") == 0)
		  {
		      prefs_set_path (i, "xmms %s");
		  }
	      }
	  }
	  else if(g_ascii_strcasecmp (line, "play_now_path") == 0)
	  {
	      if (strcmp (arg, "xmms -p %s") == 0)
	      {
		  prefs_set_path (PATH_PLAY_NOW, "xmms %s");
	      }
	      else
	      {
		  prefs_set_path (PATH_PLAY_NOW, arg);
	      }
	  }
	  else if(g_ascii_strcasecmp (line, "play_enqueue_path") == 0)
	  {
	      prefs_set_path (PATH_PLAY_ENQUEUE, arg);
	  }
	  else if(g_ascii_strcasecmp (line, "mp3gain_path") == 0)
	  {
	      prefs_set_path (PATH_MP3GAIN, arg);
	  }
	  else if(g_ascii_strcasecmp (line, "time_format") == 0)
	  {
	      /* removed in 0.87 */
	  }
	  else if((g_ascii_strcasecmp (line, "filename_format") == 0) ||
		  (g_ascii_strcasecmp (line, "export_template") == 0))
	  {  /* changed to "export_template" in 0.73CVS */
	      /* this "funky" string was the result of a wrong
		 autoconvert -- just ignore it */
	      if (strcmp (arg, "%a - %a/%T - %T.mp3") != 0)
	      {
		  if (cfg->version < 0.72)
		  {
		      /* changed the meaning of the %x in export_template */
		      gchar *sp = arg;
		      if (sp) while (*sp)
		      {
			  if (sp[0] == '%')
			  {
			      switch (sp[1]) {
			      case 'A':
				  sp[1] = 'a';
				  break;
			      case 'd':
				  sp[1] = 'A';
				  break;
			      case 'n':
				  sp[1] = 't';
				  break;
			      case 't':
				  sp[1] = 'T';
				  break;
			      default:
				  break;
			      }
			  }
			  ++sp;
		      }
		  }
		  prefs_set_string_value (EXPORT_FILES_TPL, arg);
	      }
	  }
	  else if(g_ascii_strcasecmp (line, "charset") == 0)
	  {
		if(strlen (arg))      prefs_set_charset(arg);
	  }
	  else if(g_ascii_strcasecmp (line, "id3") == 0)
	  {
	      prefs_set_id3_write((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "id3_all") == 0)
	  {
	      /* obsoleted since 0.71 */
	  }
	  else if(g_ascii_strcasecmp (line, "id3_write_id3v24") == 0)
	  {
	      prefs_set_id3_write_id3v24((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "md5") == 0)
	  {
	      prefs_set_md5tracks((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "update_existing") == 0)
	  {
	      prefs_set_update_existing((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "block_display") == 0)
	  {
	      prefs_set_block_display((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "delete_file") == 0)
	  {
	      prefs_set_track_playlist_deletion((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "delete_database") == 0)
	  {
	      prefs_set_track_database_deletion((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "delete_local_file") == 0)
	  {
	      prefs_set_track_local_file_deletion((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "delete_playlist") == 0)
	  {
	      /* ignore -- no longer supported as of 0.61-CVS */
	  }
	  else if(g_ascii_strcasecmp (line, "sync_remove_confirm") == 0)
	  {
	      prefs_set_sync_remove_confirm((gboolean)atoi(arg));
	  }
	  else if((g_ascii_strcasecmp (line, "auto_import") == 0) ||
		  (g_ascii_strcasecmp (line, "autoimport") == 0))
	  {
	      prefs_set_autoimport((gboolean)atoi(arg));
	  }
	  else if(arg_comp (line, "st_autoselect", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_st_autoselect (i, atoi (arg));
	  }
	  else if(arg_comp (line, "st_category", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_st_category (i, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_or", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_or (i, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_rating_cond", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_cond (i, T_RATING, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_playcount_cond", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_cond (i, T_PLAYCOUNT, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_played_cond", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_cond (i, T_TIME_PLAYED, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_modified_cond", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_cond (i, T_TIME_MODIFIED, atoi (arg));
	  }
	  else if((arg_comp (line, "sp_created_cond", &off) == 0) ||
		  arg_comp (line, "sp_added_cond", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_cond (i, T_TIME_ADDED, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_rating_state", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_rating_state (i, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_playcount_low", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_playcount_low (i, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_playcount_high", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_playcount_high (i, atoi (arg));
	  }
	  else if(arg_comp (line, "sp_played_state", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_entry (i, T_TIME_PLAYED, arg);
	  }
	  else if(arg_comp (line, "sp_modified_state", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_entry (i, T_TIME_MODIFIED, arg);
	  }
	  else if((arg_comp (line, "sp_created_state", &off) == 0) ||
		  (arg_comp (line, "sp_added_state", &off) == 0))
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_entry (i, T_TIME_ADDED, arg);
	  }
	  else if(arg_comp (line, "sp_autodisplay", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_sp_autodisplay (i, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "mpl_autoselect") == 0)
	  {
	      prefs_set_mpl_autoselect((gboolean)atoi(arg));
	  }
	  else if((arg_comp (line, "tm_col_width", &off) == 0) ||
		  (g_ascii_strncasecmp (line, "sm_col_width", 12) == 0))
	  {
	      gint i = atoi (line+off);
	      prefs_set_tm_col_width (i, atoi (arg));
	  }
	  else if(arg_comp (line, "tag_autoset", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_autosettags (i, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "readtags") == 0)
	  {
	      prefs_set_readtags((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "parsetags") == 0)
	  {
	      prefs_set_parsetags((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "parsetags_overwrite") == 0)
	  {
	      prefs_set_parsetags_overwrite((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "parsetags_template") == 0)
	  {
	      prefs_set_parsetags_template(strdup(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "coverart") == 0)
	  {
	      prefs_set_coverart((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "coverart_template") == 0)
	  {
	      prefs_set_coverart_template(strdup(arg));
	  }
	  else if(arg_comp (line, "col_visible", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_col_visible (i, atoi (arg));
	  }
	  else if(arg_comp (line, "col_order", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_col_order (i, atoi (arg));
	  }
	  else if(arg_comp (line, "paned_pos_", &off) == 0)
	  {
	      gint i = atoi (line+off);
	      prefs_set_paned_pos (i, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "offline") == 0)
	  {
	      prefs_set_offline((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "sort_tab_num") == 0)
	  {
	      prefs_set_sort_tab_num(atoi(arg), FALSE);
	  }
	  else if(g_ascii_strcasecmp (line, "group_compilations") == 0)
	  {
	      prefs_set_group_compilations((gboolean)atoi(arg), FALSE);
	  }
	  else if(g_ascii_strcasecmp (line, "toolbar_style") == 0)
	  {
	      prefs_set_toolbar_style(atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "pm_autostore") == 0)
	  {
	      prefs_set_pm_autostore((gboolean)atoi(arg));
	  }
	  else if((g_ascii_strcasecmp (line, "tm_autostore") == 0) ||
		  (g_ascii_strcasecmp (line, "sm_autostore") == 0))
	  {
	      prefs_set_tm_autostore((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "pm_sort") == 0)
	  {
	      gint sort = correct_sort (atoi(arg));
	      prefs_set_pm_sort(sort);
	  }
	  else if(g_ascii_strcasecmp (line, "st_sort") == 0)
	  {
	      gint sort = correct_sort (atoi(arg));
	      prefs_set_st_sort(sort);
	  }
	  else if((g_ascii_strcasecmp (line, "tm_sort_") == 0) ||
		  (g_ascii_strcasecmp (line, "sm_sort_") == 0))
	  {
	      gint sort = correct_sort (atoi(arg));
	      prefs_set_tm_sort(sort);
	  }
	  else if((g_ascii_strcasecmp (line, "tm_sortcol") == 0) ||
		  (g_ascii_strcasecmp (line, "sm_sortcol") == 0))
	  {
	      prefs_set_tm_sortcol(atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "last_prefs_page") == 0)
	  {
	      prefs_set_last_prefs_page(atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "backups") == 0)
	  {
	      /* removed with version after 0.82-CVS */
	  }
	  else if(g_ascii_strcasecmp (line, "extended_info") == 0)
	  {
	      prefs_set_write_extended_info((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "dir_browse") == 0)
	  {
	      prefs_set_last_dir_browse(strdup(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "dir_export") == 0)
	  {
	      prefs_set_string_value (EXPORT_FILES_PATH, arg);
	  }
	  else if(g_ascii_strcasecmp (line, "show_duplicates") == 0)
	  {
	      prefs_set_show_duplicates((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "show_updated") == 0)
	  {
	      prefs_set_show_updated((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "show_non_updated") == 0)
	  {
	      prefs_set_show_non_updated((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "show_sync_dirs") == 0)
	  {
	      prefs_set_show_sync_dirs((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "sync_remove") == 0)
	  {
	      prefs_set_sync_remove((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "display_toolbar") == 0)
	  {
	      prefs_set_display_toolbar((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "display_tooltips_main") == 0)
	  {
	      prefs_set_display_tooltips_main((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "display_tooltips_prefs") == 0)
	  {
	      prefs_set_display_tooltips_prefs((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "multi_edit") == 0)
	  {
	      prefs_set_multi_edit((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "multi_edit_title") == 0)
	  {
	      prefs_set_multi_edit_title((gboolean)atoi(arg));
	  }
	  else if((g_ascii_strcasecmp (line, "not_played_track") == 0) ||
		  (g_ascii_strcasecmp (line, "not_played_song") == 0))
	  {
	      prefs_set_not_played_track((gboolean)atoi(arg));
	  }
	  else if((g_ascii_strcasecmp (line, "misc_track_nr") == 0) ||
		  (g_ascii_strcasecmp (line, "misc_song_nr") == 0))
	  {
	      prefs_set_misc_track_nr(atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "update_charset") == 0)
	  {
	      prefs_set_update_charset((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "write_charset") == 0)
	  {
	      prefs_set_write_charset((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "add_recursively") == 0)
	  {
	      prefs_set_add_recursively((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "case_sensitive") == 0)
	  {
	      prefs_set_case_sensitive((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "save_sorted_order") == 0)
	  {
	      /* ignore option -- has been deleted with 0.53 */
	  }
	  else if(g_ascii_strcasecmp (line, "size_gtkpod.x") == 0)
	  {
	      prefs_set_size_gtkpod (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_gtkpod.y") == 0)
	  {
	      prefs_set_size_gtkpod (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_cal.x") == 0)
	  {
	      prefs_set_size_cal (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_cal.y") == 0)
	  {
	      prefs_set_size_cal (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_conf_sw.x") == 0)
	  {
	      prefs_set_size_conf_sw (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_conf_sw.y") == 0)
	  {
	      prefs_set_size_conf_sw (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_conf.x") == 0)
	  {
	      prefs_set_size_conf (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_conf.y") == 0)
	  {
	      prefs_set_size_conf (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_dirbr.x") == 0)
	  {
	      prefs_set_size_dirbr (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_dirbr.y") == 0)
	  {
	      prefs_set_size_dirbr (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_prefs.x") == 0)
	  {
	      prefs_set_size_prefs (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_prefs.y") == 0)
	  {
	      prefs_set_size_prefs (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "size_info.x") == 0)
	  {
	      /* changed layout of info window between 0.72 and 0.73 */
	      if (cfg->version >= 0.73)
		  prefs_set_size_info (atoi (arg), -2);
	  }
	  else if(g_ascii_strcasecmp (line, "size_info.y") == 0)
	  {
	      /* changed layout of info window between 0.72 and 0.73 */
	      if (cfg->version >= 0.73)
		  prefs_set_size_info (-2, atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "export_check_existing") == 0)
	  {
	      prefs_set_int_value (EXPORT_FILES_CHECK_EXISTING,
				   atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "fix_path") == 0)
	  {
	      /* ignore -- wie always fix the export path (replace
	       * non-compatible chars) */
	  }
	  else if(g_ascii_strcasecmp (line, "automount") == 0)
	  {
	      prefs_set_automount (atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "info_window") == 0)
	  {
	      prefs_set_info_window (atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "write_gaintag") == 0)
	  {
	      /* ignore -- not used any more */
	  }
	  else if(g_ascii_strcasecmp (line, "unused_gboolean3") == 0)
	  {
	      prefs_set_unused_gboolean3 ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "concal_autosync") == 0)
	  {
	      prefs_set_concal_autosync ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "tmp_disable_sort") == 0)
	  {
	      prefs_set_tmp_disable_sort ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "startup_messages") == 0)
	  {
	      /* only set if no upgrade/downgrade was performed */
	      if (cfg->version == g_ascii_strtod (VERSION, NULL))
		  prefs_set_startup_messages ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "special_export_charset") == 0)
	  {
	      prefs_set_int_value (EXPORT_FILES_SPECIAL_CHARSET,
				   atoi (arg));
	  }
	  else if(g_ascii_strcasecmp (line, "mserv_use") == 0)
	  {
	      prefs_set_mserv_use ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "mserv_report_probs") == 0)
	  {
	      prefs_set_mserv_report_probs ((gboolean)atoi(arg));
	  }
	  else if(g_ascii_strcasecmp (line, "mserv_username") == 0)
	  {
	      prefs_set_mserv_username (arg);
	  }
	  else if(g_ascii_strcasecmp (line, "pc_dir") == 0)
	  {
	      prefs_set_pc_dir (arg);
	  }
          else if(g_ascii_strcasecmp (line, "pc_del_age") == 0)
          {
              prefs_set_pc_del_age ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_del_age_val") == 0)
          {
              prefs_set_pc_del_age_val (atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_del_copied") == 0)
          {
              prefs_set_pc_del_copied ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_auto_fetch") == 0)
          {
              prefs_set_pc_auto_fetch ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_log") == 0)
          {
              prefs_set_pc_log ((gboolean)atoi(arg));
          }
	  else if(g_ascii_strcasecmp (line, "pc_log_file") == 0)
	  {
	      prefs_set_pc_log_file (arg);
	  }
          else if(g_ascii_strcasecmp (line, "pc_auto_sync") == 0)
          {
              prefs_set_pc_auto_sync ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_ipod_del_age") == 0)
          {
              prefs_set_pc_ipod_del_age ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_ipod_del_age_val") == 0)
          {
              prefs_set_pc_ipod_del_age_val (atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_ipod_del_played") == 0)
          {
              prefs_set_pc_ipod_del_played ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_ipod_inc_date") == 0)
          {
              prefs_set_pc_ipod_inc_date ((gboolean)atoi(arg));
          }
          else if(g_ascii_strcasecmp (line, "pc_change_genre") == 0)
          {
              prefs_set_pc_change_genre ((gboolean)atoi(arg));
          }

	  else
	  {   /* All leftover options will be stored into the prefs
		 setting hash (generic options -- should have had this
		 idea much sooner... */
	      gboolean skip = FALSE;
	      if (cfg->version < 0.91)
	      {
		  if(arg_comp (line, "itdb_", NULL) == 0)
		  {   /* set incorrectly in 0.90 -- delete */
		      skip = TRUE;
		  }
	      }
	      if (!skip)
		  prefs_set_string_value (line, arg);
	  }
	  g_free(line);
	}
    }
}


/* we first read from /etc/gtkpod/prefs and then overwrite the
   settings with ~/.gtkpod/prefs */
void
read_prefs_defaults(void)
{
  gchar *cfgdir = NULL;
  gchar *filename;
  FILE *fp = NULL;
  gboolean have_prefs = FALSE;

  cfgdir = prefs_get_cfgdir ();

  filename = g_build_filename (cfgdir, "prefs", NULL);
  if(g_file_test(filename, G_FILE_TEST_EXISTS))
  {
      if((fp = fopen(filename, "r")))
      {
	  read_prefs_from_file_desc(fp);
	  fclose(fp);
	  have_prefs = TRUE; /* read prefs */
      }
      else
      {
	  gtkpod_warning(_("Unable to open config file '%s' for reading\n"), filename);
	  }
  }
  g_free (filename);

  if (!have_prefs)
  {
      filename = g_build_filename ("/etc", "gtkpod", "prefs", NULL);

      if (g_file_test (filename, G_FILE_TEST_EXISTS))
      {
	  if((fp = fopen(filename, "r")))
	  {
	      read_prefs_from_file_desc(fp);
	      fclose(fp);
	      have_prefs = TRUE; /* read prefs */
	  }
      }
      g_free (filename);
  }
  g_free (cfgdir);
  /* set version of the prefs file to "current" if none was read */
  if (!have_prefs)   cfg->version = g_ascii_strtod (VERSION, NULL);

  /* handle version changes in prefs */
  if (cfg->version == 0.0)
  {
      /* most likely prefs file written by V0.50 */
      /* I added two new PANED elements since V0.50 --> shift */
      gint i;
      for (i=PANED_NUM_ST-1; i>=0; --i)
      {
	  prefs_set_paned_pos (PANED_NUM_GLADE + i,
			       prefs_get_paned_pos (PANED_NUM_GLADE + i - 2));
      }
      prefs_set_paned_pos (PANED_STATUS1, -1);
      prefs_set_paned_pos (PANED_STATUS2, -1);
  }

  /* set statusbar paned to a decent value if unset */
  if (prefs_get_paned_pos (PANED_STATUS1) == -1)
  {
      gint x,y;
      prefs_get_size_gtkpod (&x, &y);
      /* set to about 2/3 of the window width */
      if (x>0)   prefs_set_paned_pos (PANED_STATUS1, 20*x/30);
  }
  /* set statusbar paned to a decent value if unset */
  if (prefs_get_paned_pos (PANED_STATUS2) == -1)
  {
      gint x,y,p;
      prefs_get_size_gtkpod (&x, &y);
      p = prefs_get_paned_pos (PANED_STATUS1);
      /* set to about half of the remaining window */
      if (x>0)   prefs_set_paned_pos (PANED_STATUS2, (x-p)/2 );
  }
}

/* Read Preferences and initialise the cfg-struct */
/* Return value: FALSE if "-p" argument was given -> stop program */
gboolean read_prefs (GtkWidget *gtkpod, int argc, char *argv[])
{
  GtkCheckMenuItem *menu;
  int opt;
  int option_index;
  gboolean result = TRUE;
  struct option const options[] =
    {
      { "h",           no_argument,	NULL, GP_HELP },
      { "help",        no_argument,	NULL, GP_HELP },
      { "p",           required_argument,       NULL, GP_PLAYCOUNT },
      { "m",           required_argument,	NULL, GP_MOUNT },
      { "mountpoint",  required_argument,	NULL, GP_MOUNT },
      { "offline",     no_argument,	NULL, GP_OFFLINE },
      { "a",           no_argument,	NULL, GP_AUTO },
      { "auto",        no_argument,	NULL, GP_AUTO },
      { 0, 0, 0, 0 }
    };

  if (cfg != NULL) discard_prefs ();

  cfg = cfg_new();
  read_prefs_defaults();

  prefs_set_ipod_mount (getenv("IPOD_MOUNTPOINT"));

  while((opt=getopt_long_only(argc, argv, "", options, &option_index)) != -1) {
    switch(opt)
      {
      case GP_HELP:
	  usage(stdout);
	  exit(0);
	  break;
      case GP_PLAYCOUNT:
	  client_playcount (optarg);
	  result = FALSE;
	  break;
      case GP_MOUNT:
	  prefs_set_ipod_mount (optarg);
	  break;
      case GP_AUTO:
	  prefs_set_autoimport_commandline (TRUE);
	  break;
      case GP_OFFLINE:
	  prefs_set_offline (TRUE);
	  break;
      default:
	  usage_fpf(stderr, _("Unknown option: %s\n"), argv[optind]);
	  usage(stderr);
	  exit(1);
      }
  }

/* FIXME: PODCASTS: remove Podcast menu */
/*   podcast_read_from_file(); */
/*   podcast_file_read_from_file(); */

  menu = GTK_CHECK_MENU_ITEM (gtkpod_xml_get_widget (main_window_xml, "offline_menu"));
  gtk_check_menu_item_set_active (menu, prefs_get_offline ());
  return result;
}

static void
write_prefs_to_file_desc(FILE *fp)
{
    gint i;

    if(!fp)
	fp = stderr;

    /* update column widths, x,y-size of main window and GtkPaned
     * positions */
    display_update_default_sizes ();
    /* update order of track view columns */
    tm_store_col_order ();

    fprintf(fp, "version=%s\n", VERSION);
    fprintf(fp, "mountpoint=%s\n", cfg->ipod_mount);
    for (i=0; i<PATH_NUM; ++i)
    {
	gchar *buf = g_strdup (path_entry_names[i]);
	gchar *bufp = strrchr (buf, '_');
	/* we cut off the "_entry" of the entry name, making it easier
	   to read for humans */
	if (bufp) *bufp = '\0';
	fprintf (fp, ";%s\n", buf);
	fprintf (fp, "path%d=%s\n", i, cfg->path[i]);
	g_free (buf);
    }
    if (cfg->charset)
    {
	fprintf(fp, "charset=%s\n", cfg->charset);
    } else {
	fprintf(fp, "charset=\n");
    }
    fprintf(fp, "id3=%d\n", prefs_get_id3_write ());
    fprintf(fp, "id3_write_id3v24=%d\n", prefs_get_id3_write_id3v24 ());
    fprintf(fp, "md5=%d\n",prefs_get_md5tracks ());
    fprintf(fp, "update_existing=%d\n",prefs_get_update_existing ());
    fprintf(fp, "block_display=%d\n",prefs_get_block_display());
    fprintf(fp, _("# delete confirmation\n"));
    fprintf(fp, "delete_file=%d\n",prefs_get_track_playlist_deletion());
    fprintf(fp, "delete_ipod=%d\n",prefs_get_track_ipod_file_deletion());
    fprintf(fp, "delete_database=%d\n",prefs_get_track_database_deletion());
    fprintf(fp, "delete_local_file=%d\n",prefs_get_track_local_file_deletion());
    fprintf(fp, "sync_remove_confirm=%d\n",prefs_get_sync_remove_confirm());
    fprintf(fp, "autoimport=%d\n",prefs_get_autoimport());
    fprintf(fp, _("# sort tab: select 'All', last selected page (=category)\n"));
    for (i=0; i<SORT_TAB_MAX; ++i)
    {
	fprintf(fp, "st_autoselect%d=%d\n", i, prefs_get_st_autoselect (i));
	fprintf(fp, "st_category%d=%d\n", i, prefs_get_st_category (i));
	fprintf(fp, "sp_or%d=%d\n", i, prefs_get_sp_or (i));
	fprintf(fp, "sp_rating_cond%d=%d\n", i, prefs_get_sp_cond (i, T_RATING));
	fprintf(fp, "sp_rating_state%d=%d\n", i, prefs_get_sp_rating_state(i));
	fprintf(fp, "sp_playcount_cond%d=%d\n", i, prefs_get_sp_cond (i, T_PLAYCOUNT));
	fprintf(fp, "sp_playcount_low%d=%d\n", i, prefs_get_sp_playcount_low (i));
	fprintf(fp, "sp_playcount_high%d=%d\n", i, prefs_get_sp_playcount_high (i));
	fprintf(fp, "sp_played_cond%d=%d\n", i, prefs_get_sp_cond (i, T_TIME_PLAYED));
	fprintf(fp, "sp_played_state%d=%s\n", i, prefs_get_sp_entry (i, T_TIME_PLAYED));
	fprintf(fp, "sp_modified_cond%d=%d\n", i, prefs_get_sp_cond (i, T_TIME_MODIFIED));
	fprintf(fp, "sp_modified_state%d=%s\n", i, prefs_get_sp_entry (i, T_TIME_MODIFIED));
	fprintf(fp, "sp_added_cond%d=%d\n", i, prefs_get_sp_cond (i, T_TIME_ADDED));
	fprintf(fp, "sp_added_state%d=%s\n", i, prefs_get_sp_entry (i, T_TIME_ADDED));
	fprintf(fp, "sp_autodisplay%d=%d\n", i, prefs_get_sp_autodisplay (i));
    }
    fprintf(fp, _("# autoselect master playlist?\n"));
    fprintf(fp, "mpl_autoselect=%d\n", prefs_get_mpl_autoselect ());
    fprintf(fp, _("# title=0, artist, album, genre, composer\n"));
    fprintf(fp, _("# track_nr=5, ipod_id, pc_path, transferred\n"));
    fprintf(fp, _("# autoset: set empty tag to filename?\n"));
    for (i=0; i<TM_NUM_COLUMNS; ++i)
    {
	fprintf(fp, "tm_col_width%d=%d\n", i, prefs_get_tm_col_width (i));
	fprintf(fp, "col_visible%d=%d\n",  i, prefs_get_col_visible (i));
	fprintf(fp, "col_order%d=%d\n",  i, prefs_get_col_order (i));
	if (i < TM_NUM_TAGS_PREFS)
	    fprintf(fp, "tag_autoset%d=%d\n", i, prefs_get_autosettags (i));
    }
    fprintf(fp, "readtags=%d\n", prefs_get_readtags());
    fprintf(fp, "parsetags=%d\n", prefs_get_parsetags());
    fprintf(fp, "parsetags_overwrite=%d\n", prefs_get_parsetags_overwrite());
    fprintf(fp, "parsetags_template=%s\n",cfg->parsetags_template);
    fprintf(fp, "coverart=%d\n", prefs_get_coverart());
    fprintf(fp, "coverart_template=%s\n",cfg->coverart_template);
    fprintf(fp, _("# position of sliders (paned): playlists, above tracks,\n# between sort tabs, and in statusbar.\n"));
    for (i=0; i<PANED_NUM; ++i)
    {
	fprintf(fp, "paned_pos_%d=%d\n", i, prefs_get_paned_pos (i));
    }
    fprintf(fp, "sort_tab_num=%d\n",prefs_get_sort_tab_num());
    fprintf(fp, "group_compilations=%d\n",prefs_get_group_compilations());
    fprintf(fp, "last_prefs_page=%d\n",prefs_get_last_prefs_page());
    fprintf(fp, "offline=%d\n",prefs_get_offline());
    fprintf(fp, "extended_info=%d\n",prefs_get_write_extended_info());
    fprintf(fp, "dir_browse=%s\n",cfg->last_dir.browse);
    fprintf(fp, "show_duplicates=%d\n",prefs_get_show_duplicates());
    fprintf(fp, "show_updated=%d\n",prefs_get_show_updated());
    fprintf(fp, "show_non_updated=%d\n",prefs_get_show_non_updated());
    fprintf(fp, "show_sync_dirs=%d\n",prefs_get_show_sync_dirs());
    fprintf(fp, "sync_remove=%d\n",prefs_get_sync_remove());
    fprintf(fp, "display_toolbar=%d\n",prefs_get_display_toolbar());
    fprintf(fp, "toolbar_style=%d\n",prefs_get_toolbar_style());
    fprintf(fp, "pm_autostore=%d\n",prefs_get_pm_autostore());
    fprintf(fp, "tm_autostore=%d\n",prefs_get_tm_autostore());
    fprintf(fp, "pm_sort=%d\n",prefs_get_pm_sort());
    fprintf(fp, "st_sort=%d\n",prefs_get_st_sort());
    fprintf(fp, "tm_sort_=%d\n",prefs_get_tm_sort());
    fprintf(fp, "tm_sortcol=%d\n",prefs_get_tm_sortcol());
    fprintf(fp, "display_tooltips_main=%d\n",
	    prefs_get_display_tooltips_main());
    fprintf(fp, "display_tooltips_prefs=%d\n",
	    prefs_get_display_tooltips_prefs());
    fprintf(fp, "multi_edit=%d\n", prefs_get_multi_edit());
    fprintf(fp, "multi_edit_title=%d\n", prefs_get_multi_edit_title());
    fprintf(fp, "misc_track_nr=%d\n", prefs_get_misc_track_nr());
    fprintf(fp, "not_played_track=%d\n", prefs_get_not_played_track());
    fprintf(fp, "update_charset=%d\n",prefs_get_update_charset());
    fprintf(fp, "write_charset=%d\n",prefs_get_write_charset());
    fprintf(fp, "add_recursively=%d\n",prefs_get_add_recursively());
    fprintf(fp, "case_sensitive=%d\n",prefs_get_case_sensitive());
    fprintf(fp, _("# window sizes: main window, confirmation scrolled,\n#               confirmation non-scrolled, dirbrowser, prefs\n"));
    fprintf (fp, "size_gtkpod.x=%d\n", cfg->size_gtkpod.x);
    fprintf (fp, "size_gtkpod.y=%d\n", cfg->size_gtkpod.y);
    fprintf (fp, "size_cal.x=%d\n", cfg->size_cal.x);
    fprintf (fp, "size_cal.y=%d\n", cfg->size_cal.y);
    fprintf (fp, "size_conf_sw.x=%d\n", cfg->size_conf_sw.x);
    fprintf (fp, "size_conf_sw.y=%d\n", cfg->size_conf_sw.y);
    fprintf (fp, "size_conf.x=%d\n", cfg->size_conf.x);
    fprintf (fp, "size_conf.y=%d\n", cfg->size_conf.y);
    fprintf (fp, "size_dirbr.x=%d\n", cfg->size_dirbr.x);
    fprintf (fp, "size_dirbr.y=%d\n", cfg->size_dirbr.y);
    fprintf (fp, "size_prefs.x=%d\n", cfg->size_prefs.x);
    fprintf (fp, "size_prefs.y=%d\n", cfg->size_prefs.y);
    fprintf (fp, "size_info.x=%d\n", cfg->size_info.x);
    fprintf (fp, "size_info.y=%d\n", cfg->size_info.y);
    fprintf (fp, "automount=%d\n", cfg->automount);
    fprintf (fp, "info_window=%d\n", cfg->info_window);
    fprintf (fp, "concal_autosync=%d\n", cfg->concal_autosync);
    prefs_write_hash_values (fp);
    fprintf (fp, "tmp_disable_sort=%d\n", cfg->tmp_disable_sort);
    fprintf (fp, "startup_messages=%d\n", cfg->startup_messages);
    fprintf (fp, "mserv_use=%d\n", cfg->mserv_use);
    fprintf (fp, "mserv_report_probs=%d\n", cfg->mserv_report_probs);
    fprintf(fp, "pc_dir=%s\n", cfg->pc_dir);
    fprintf(fp, "pc_del_age=%d\n", cfg->pc_del_age);
    fprintf(fp, "pc_del_age_val=%d\n", cfg->pc_del_age_val);
    fprintf(fp, "pc_del_copied=%d\n", cfg->pc_del_copied);
    fprintf(fp, "pc_auto_fetch=%d\n", cfg->pc_auto_fetch);
    fprintf(fp, "pc_log=%d\n", cfg->pc_log);
    fprintf(fp, "pc_log_file=%s\n", cfg->pc_log_file);
    fprintf(fp, "pc_auto_sync=%d\n", cfg->pc_auto_sync);
    fprintf(fp, "pc_ipod_del_age=%d\n", cfg->pc_ipod_del_age);
    fprintf(fp, "pc_ipod_del_age_val=%d\n", cfg->pc_ipod_del_age_val);
    fprintf(fp, "pc_ipod_del_played=%d\n", cfg->pc_ipod_del_played);
    fprintf(fp, "pc_ipod_inc_date=%d\n", cfg->pc_ipod_inc_date);
    fprintf(fp, "pc_change_genre=%d\n", cfg->pc_change_genre);
/*     fprintf (fp, "unused_gboolean3=%d\n", cfg->unused_gboolean3); */
}


void
write_prefs (void)
{
    gchar *filename;
    gchar *cfgdir;
    FILE *fp = NULL;

    cfgdir = prefs_get_cfgdir ();

    filename = g_build_filename (cfgdir, "prefs", NULL);
    if((fp = fopen(filename, "w")))
    {
	write_prefs_to_file_desc(fp);
	fclose(fp);
    }
    else
    {
	gtkpod_warning (_("Unable to open '%s' for writing\n"),
			filename);
    }

    g_free (filename);
    g_free (cfgdir);
}



/* Free all memory including the cfg-struct itself. */
void discard_prefs ()
{
    cfg_free(cfg);
    cfg = NULL;
}

void cfg_free(struct cfg *c)
{
    if(c)
    {
      gint i;

      g_free (c->ipod_mount);
      g_free (c->charset);
      g_free (c->last_dir.browse);
      for (i=0; i<PATH_NUM; ++i)
	  g_free (c->path[i]);
      g_free (c->mserv_username);
      g_free (c);
    }
}

void sortcfg_free(struct sortcfg *c)
{
    g_return_if_fail (c);
    g_list_free (c->tmp_sort_ign_fields);
    g_free (c->tmp_sort_ign_strings);
    g_free (c);
}


static gchar *
get_dirname_of_filename(const gchar *file)
{
    if (!file) return NULL;

    if (g_file_test(file, G_FILE_TEST_IS_DIR))
	return g_strdup (file);
    else
	return g_path_get_dirname (file);
}


void prefs_set_offline(gboolean active)
{
  if (cfg->offline != active)   space_data_update ();
  cfg->offline = active;
  info_update_totals_view_space ();
}
void prefs_set_write_extended_info(gboolean active)
{
  cfg->write_extended_info = active;
}
void prefs_set_last_dir_browse(const gchar *file)
{
    if (file)
    {
	C_FREE(cfg->last_dir.browse);
	cfg->last_dir.browse = get_dirname_of_filename(file);
    }
}
const gchar *prefs_get_last_dir_browse(void)
{
    return cfg->last_dir.browse;
}

void prefs_set_ipod_mount(const gchar *mp)
{
    if (!mp) return; /* ignore NULL pointer */
    g_free (cfg->ipod_mount);
    /* if new mount point starts with "~/", we replace it with the
       home directory */
    cfg->ipod_mount = convert_filename (mp);
    /* need to notify the info thread of new mount point */
    space_set_ipod_mount (cfg->ipod_mount);
    gp_itdb_set_mountpoint (cfg->ipod_mount);
}


/* If the status of md5 hash flag changes, free or re-init the md5
   hash table */
void prefs_set_md5tracks (gboolean active)
{
    struct itdbs_head *itdbs_head;

    g_return_if_fail (gtkpod_window);
    itdbs_head = g_object_get_data (G_OBJECT (gtkpod_window),
				    "itdbs_head");
/*    g_return_if_fail (itdbs_head);*/
    /* gets called before itdbs are set up -> fail silently */
    if (!itdbs_head)
    {
	cfg->md5tracks = active;
	return;
    }

    if (cfg->md5tracks && !active)
    { /* md5 checksums have been turned off */
	cfg->md5tracks = FALSE;
	gp_md5_free_hash ();
    }
    if (!cfg->md5tracks && active)
    { /* md5 checksums have been turned on */
	cfg->md5tracks = TRUE; /* must be set before calling
				  gp_md5_hash_tracks() */
	gp_md5_hash_tracks ();
	/* display duplicates */
	gp_duplicate_remove (NULL, NULL);
    }
}

gboolean prefs_get_md5tracks(void)
{
    return cfg->md5tracks;
}

void prefs_set_update_existing(gboolean active)
{
    cfg->update_existing = active;
}

gboolean prefs_get_update_existing(void)
{
    return cfg->update_existing;
}

/* Should the display be blocked (be insenstive) while it is updated
   after a selection (playlist, tab entry) change? */
void prefs_set_block_display(gboolean active)
{
    cfg->block_display = active;
}

gboolean prefs_get_block_display(void)
{
    return cfg->block_display;
}

void prefs_set_id3_write(gboolean active)
{
    cfg->id3_write = active;
}

gboolean prefs_get_id3_write(void)
{
    return cfg->id3_write;
}

void prefs_set_id3_write_id3v24(gboolean active)
{
    cfg->id3_write_id3v24 = active;
}

gboolean prefs_get_id3_write_id3v24(void)
{
    return cfg->id3_write_id3v24;
}

gboolean prefs_get_offline(void)
{
  return cfg->offline;
}

gboolean prefs_get_write_extended_info(void)
{
  return cfg->write_extended_info;
}

void prefs_set_track_playlist_deletion(gboolean val)
{
    cfg->deletion.track = val;
}

gboolean prefs_get_track_playlist_deletion(void)
{
    return(cfg->deletion.track);
}

void prefs_set_track_ipod_file_deletion(gboolean val)
{
    cfg->deletion.ipod_file = val;
}

gboolean prefs_get_track_ipod_file_deletion(void)
{
    return(cfg->deletion.ipod_file);
}

void prefs_set_track_database_deletion(gboolean val)
{
    cfg->deletion.database = val;
}

gboolean prefs_get_track_database_deletion(void)
{
    return (cfg->deletion.database);
}

void prefs_set_track_local_file_deletion(gboolean val)
{
    cfg->deletion.local_file = val;
}

gboolean prefs_get_track_local_file_deletion(void)
{
    return cfg->deletion.local_file;
}


void prefs_set_sync_remove_confirm(gboolean val)
{
    cfg->deletion.syncing = val;
}

gboolean prefs_get_sync_remove_confirm(void)
{
    return(cfg->deletion.syncing);
}

void prefs_set_charset (gchar *charset)
{
    prefs_cfg_set_charset (cfg, charset);
}


void prefs_cfg_set_charset (struct cfg *cfgd, gchar *charset)
{
    C_FREE (cfgd->charset);
    if (charset && strlen (charset))
	cfgd->charset = g_strdup (charset);
/*     printf ("set_charset: '%s'\n", charset);	 */
}


gchar *prefs_get_charset (void)
{
    return cfg->charset;
/*     printf ("get_charset: '%s'\n", cfg->charset); */
}

struct cfg *clone_prefs(void)
{
    struct cfg *result = NULL;

    if(cfg)
    {
	gint i;

	result = g_memdup (cfg, sizeof (struct cfg));
	result->ipod_mount = g_strdup(cfg->ipod_mount);
	result->charset = g_strdup(cfg->charset);
	result->last_dir.browse = g_strdup(cfg->last_dir.browse);
	for (i=0; i<PATH_NUM; ++i)
	    result->path[i] = g_strdup (cfg->path[i]);
	result->parsetags_template = g_strdup(cfg->parsetags_template);
	result->coverart_template = g_strdup(cfg->coverart_template);
	result->mserv_username = g_strdup(cfg->mserv_username);
	result->pc_dir = g_strdup(cfg->pc_dir);
	result->pc_log_file = g_strdup(cfg->pc_log_file);
    }
    return(result);
}

struct sortcfg *clone_sortprefs(void)
{
    struct sortcfg *result = NULL;

    if(cfg)
    {
	result = g_memdup (&cfg->sortcfg, sizeof (struct sortcfg));
    }
    /* GLists are not copied */
    result->tmp_sort_ign_fields = NULL;
    result->tmp_sort_ign_strings = NULL;
    return(result);
}

gboolean prefs_get_autoimport(void)
{
    return(cfg->autoimport);
}

void prefs_set_autoimport(gboolean val)
{
    cfg->autoimport = val;
}

gboolean prefs_get_autoimport_commandline(void)
{
    return(cfg->autoimport_commandline);
}

void prefs_set_autoimport_commandline(gboolean val)
{
    cfg->autoimport_commandline = val;
}

/* "inst": the instance of the sort tab */
gboolean prefs_get_st_autoselect (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
    {
	return cfg->st[inst].autoselect;
    }
    else
    {
	return TRUE;
    }
}

/* "inst": the instance of the sort tab, "autoselect": should "All" be
 * selected automatically? */
void prefs_set_st_autoselect (guint32 inst, gboolean autoselect)
{
    if (inst < SORT_TAB_MAX)
    {
	cfg->st[inst].autoselect = autoselect;
    }
}

/* "inst": the instance of the sort tab */
guint prefs_get_st_category (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
    {
	return cfg->st[inst].category;
    }
    else
    {
	return TRUE; /* hmm.... this should not happen... */
    }
}

/* "inst": the instance of the sort tab, "category": one of the
   ST_CAT_... */
void prefs_set_st_category (guint32 inst, guint category)
{
    if ((inst < SORT_TAB_MAX) && (category < ST_CAT_NUM))
    {
	cfg->st[inst].category = category;
    }
    else
    {
	gtkpod_warning (_(" Preferences: Category nr (%d<%d?) or sorttab nr (%d<%d?) out of range.\n"), category, ST_CAT_NUM, inst, SORT_TAB_MAX);
    }
}

gboolean prefs_get_mpl_autoselect (void)
{
    return cfg->mpl_autoselect;
}

/* Should the MPL be selected automatically? */
void prefs_set_mpl_autoselect (gboolean autoselect)
{
    cfg->mpl_autoselect = autoselect;
}


/* retrieve the width of the track display columns. "col": one of the
   TM_COLUMN_... */
gint prefs_get_tm_col_width (gint col)
{
    if (col < TM_NUM_COLUMNS && (cfg->tm_col_width[col] > 0))
	return cfg->tm_col_width[col];
    return 80;  /* default -- col should be smaller than
		   TM_NUM_COLUMNS) */
}

/* set the width of the track display columns. "col": one of the
   TM_COLUMN_..., "width": current width */
void prefs_set_tm_col_width (gint col, gint width)
{
    if (col < TM_NUM_COLUMNS && width > 0)
	cfg->tm_col_width[col] = width;
}


/* Returns "$HOME/.gtkpod" and tries to create it if it does not
   exist. */
gchar *prefs_get_cfgdir (void)
{
  G_CONST_RETURN gchar *str;
  gchar *cfgdir=NULL;

  if((str = g_get_home_dir ()))
  {
      cfgdir = g_build_filename (str, ".gtkpod", NULL);
      if(!g_file_test(cfgdir, G_FILE_TEST_IS_DIR))
      {
	  if(mkdir(cfgdir, 0777) == -1)
	  {
	      gtkpod_warning(_("Unable to 'mkdir %s'\n"), cfgdir);
	  }
      }
  }
  return cfgdir;
}

/* Returns the ipod_mount. Don't modify it! */
const gchar *prefs_get_ipod_mount (void)
{
    return cfg->ipod_mount;
}

/* Sets the default size for the gtkpod window. -2 means: don't change
 * the current size */
void prefs_set_size_gtkpod (gint x, gint y)
{
    if (x != -2) cfg->size_gtkpod.x = x;
    if (y != -2) cfg->size_gtkpod.y = y;
}

/* Sets the default size for the calendar window. -2 means: don't change
 * the current size */
void prefs_set_size_cal (gint x, gint y)
{
    if (x != -2) cfg->size_cal.x = x;
    if (y != -2) cfg->size_cal.y = y;
}

/* Sets the default size for the scrolled conf window. -2 means: don't
 * change the current size */
void prefs_set_size_conf_sw (gint x, gint y)
{
    if (x != -2) cfg->size_conf_sw.x = x;
    if (y != -2) cfg->size_conf_sw.y = y;
}

/* Sets the default size for the non-scrolled conf window. -2 means:
 * don't change the current size */
void prefs_set_size_conf (gint x, gint y)
{
    if (x != -2) cfg->size_conf.x = x;
    if (y != -2) cfg->size_conf.y = y;
}

/* Sets the default size for the dirbrowser window. -2 means:
 * don't change the current size */
void prefs_set_size_dirbr (gint x, gint y)
{
    if (x != -2) cfg->size_dirbr.x = x;
    if (y != -2) cfg->size_dirbr.y = y;
}

/* Sets the default size for the prefs window. -2 means:
 * don't change the current size */
void prefs_set_size_prefs (gint x, gint y)
{
    if (x != -2) cfg->size_prefs.x = x;
    if (y != -2) cfg->size_prefs.y = y;
}

/* Sets the default size for the info window. -2 means:
 * don't change the current size */
void prefs_set_size_info (gint x, gint y)
{
    if (x != -2) cfg->size_info.x = x;
    if (y != -2) cfg->size_info.y = y;
}

/* Writes the current default size for the gtkpod window in "x" and
   "y" */
void prefs_get_size_gtkpod (gint *x, gint *y)
{
    *x = cfg->size_gtkpod.x;
    *y = cfg->size_gtkpod.y;
}

/* Writes the current default size for the gtkpod window in "x" and
   "y" */
void prefs_get_size_cal (gint *x, gint *y)
{
    *x = cfg->size_cal.x;
    *y = cfg->size_cal.y;
}

/* Writes the current default size for the scrolled conf window in "x"
   and "y" */
void prefs_get_size_conf_sw (gint *x, gint *y)
{
    *x = cfg->size_conf_sw.x;
    *y = cfg->size_conf_sw.y;
}

/* Writes the current default size for the non-scrolled conf window in
   "x" and "y" */
void prefs_get_size_conf (gint *x, gint *y)
{
    *x = cfg->size_conf.x;
    *y = cfg->size_conf.y;
}


/* Writes the current default size for the dirbrowser window in
   "x" and "y" */
void prefs_get_size_dirbr (gint *x, gint *y)
{
    *x = cfg->size_dirbr.x;
    *y = cfg->size_dirbr.y;
}


/* Writes the current default size for the info window in
   "x" and "y" */
void prefs_get_size_prefs (gint *x, gint *y)
{
    *x = cfg->size_prefs.x;
    *y = cfg->size_prefs.y;
}

/* Writes the current default size for the info window in
   "x" and "y" */
void prefs_get_size_info (gint *x, gint *y)
{
    *x = cfg->size_info.x;
    *y = cfg->size_info.y;
}


/* Should empty tags be set to filename? -- "category": one of the
   TM_COLUMN_..., "autoset": new value */
void prefs_set_autosettags (gint category, gboolean autoset)
{
    if (category < TM_NUM_TAGS_PREFS)
	cfg->autosettags[category] = autoset;
}


/* Should empty tags be set to filename? -- "category": one of the
   TM_COLUMN_... */
gboolean prefs_get_autosettags (gint category)
{
    if (category < TM_NUM_TAGS_PREFS)
	return cfg->autosettags[category];
    return FALSE;
}

void prefs_set_readtags(gboolean active)
{
  cfg->readtags = active;
}

gboolean prefs_get_readtags(void)
{
  return cfg->readtags;
}

void prefs_set_parsetags(gboolean active)
{
  cfg->parsetags = active;
}

gboolean prefs_get_parsetags(void)
{
  return cfg->parsetags;
}

void prefs_set_parsetags_overwrite(gboolean active)
{
  cfg->parsetags_overwrite = active;
}

gboolean prefs_get_parsetags_overwrite(void)
{
  return cfg->parsetags_overwrite;
}

const gchar *prefs_get_parsetags_template (void)
{
    return cfg->parsetags_template;
}

void prefs_set_parsetags_template (const gchar *tpl)
{
    if (tpl)
    {
	g_free(cfg->parsetags_template);
	cfg->parsetags_template = g_strdup (tpl);
    }
}

void prefs_set_coverart(gboolean active)
{
  cfg->coverart = active;
}

gboolean prefs_get_coverart(void)
{
  return cfg->coverart;
}

const gchar *prefs_get_coverart_template (void)
{
    return cfg->coverart_template;
}

void prefs_set_coverart_template (const gchar *tpl)
{
    if (tpl)
    {
	g_free(cfg->coverart_template);
	cfg->coverart_template = g_strdup (tpl);
    }
}

/* Display column tm_item @visible: new value */
void prefs_set_col_visible (TM_item tm_item, gboolean visible)
{
    if (tm_item < TM_NUM_COLUMNS)
	cfg->col_visible[tm_item] = visible;
}


/* Display column tm_item? */
gboolean prefs_get_col_visible (TM_item tm_item)
{
    if (tm_item < TM_NUM_COLUMNS)
	return cfg->col_visible[tm_item];
    return FALSE;
}

/* Display which column at nr @pos? */
void prefs_set_col_order (gint pos, TM_item tm_item)
{
    if (pos < TM_NUM_COLUMNS)
	cfg->col_order[pos] = tm_item;
}


/* Display column nr @pos? */
TM_item prefs_get_col_order (gint pos)
{
    if (pos < TM_NUM_COLUMNS)
	return cfg->col_order[pos];
    return -1;
}

/* get position of GtkPaned element nr. "i" */
/* return value: -1: don't change position */
gint prefs_get_paned_pos (gint i)
{
    if (i < PANED_NUM)
	return cfg->paned_pos[i];
    else
    {
	g_warning ("Programming error: prefs_get_paned_pos: arg out of range (%d)\n", i);
	return 100; /* something reasonable? */
    }
}

/* set position of GtkPaned element nr. "i" */
void prefs_set_paned_pos (gint i, gint pos)
{
    if (i < PANED_NUM)
	cfg->paned_pos[i] = pos;
}


/* A value of "0" will set the default defined in misc.c */
void prefs_set_statusbar_timeout (guint32 val)
{
    if (val == 0)  val = STATUSBAR_TIMEOUT;
    cfg->statusbar_timeout = val;
    gtkpod_statusbar_reset_timeout ();
}

guint32 prefs_get_statusbar_timeout (void)
{
    return cfg->statusbar_timeout;
}

gboolean prefs_get_show_duplicates (void)
{
    return cfg->show_duplicates;
}

void prefs_set_show_duplicates (gboolean val)
{
    cfg->show_duplicates = val;
}

gboolean prefs_get_show_updated (void)
{
    return cfg->show_updated;
}

void prefs_set_show_updated (gboolean val)
{
    cfg->show_updated = val;
}

gboolean prefs_get_show_non_updated (void)
{
    return cfg->show_non_updated;
}

void prefs_set_show_non_updated (gboolean val)
{
    cfg->show_non_updated = val;
}

gboolean prefs_get_show_sync_dirs (void)
{
    return cfg->show_sync_dirs;
}

void prefs_set_show_sync_dirs (gboolean val)
{
    cfg->show_sync_dirs = val;
}

gboolean prefs_get_sync_remove (void)
{
    return cfg->sync_remove;
}

void prefs_set_sync_remove (gboolean val)
{
    cfg->sync_remove = val;
}

gboolean prefs_get_display_toolbar (void)
{
    return cfg->display_toolbar;
}

void prefs_set_display_toolbar (gboolean val)
{
    cfg->display_toolbar = val;
    display_show_hide_toolbar ();
}

gboolean prefs_get_update_charset (void)
{
    return cfg->update_charset;
}

void prefs_set_update_charset (gboolean val)
{
    cfg->update_charset = val;
}

gboolean prefs_get_write_charset (void)
{
    return cfg->write_charset;
}

void prefs_set_write_charset (gboolean val)
{
    cfg->write_charset = val;
}

gboolean prefs_get_add_recursively (void)
{
    return cfg->add_recursively;
}

void prefs_set_add_recursively (gboolean val)
{
    cfg->add_recursively = val;
}

gboolean prefs_get_case_sensitive (void)
{
    return cfg->sortcfg.case_sensitive;
}

void prefs_set_case_sensitive (gboolean val)
{
    cfg->sortcfg.case_sensitive = val;
}

gint prefs_get_sort_tab_num (void)
{
    return cfg->sort_tab_num;
}

void prefs_set_sort_tab_num (gint i, gboolean update_display)
{
    if ((i>=0) && (i<=SORT_TAB_MAX))
    {
	if (cfg->sort_tab_num != i)
	{
	    cfg->sort_tab_num = i;
	    if (update_display) st_show_visible ();
	}
    }
}

gboolean prefs_get_group_compilations (void)
{
    return cfg->group_compilations;
}

void prefs_set_group_compilations (gboolean val, gboolean update_display)
{
    cfg->group_compilations = val;
    if (update_display) st_show_visible ();
}

GtkToolbarStyle prefs_get_toolbar_style (void)
{
    return cfg->toolbar_style;
}

void prefs_set_toolbar_style (GtkToolbarStyle i)
{
    switch (i)
    {
    case GTK_TOOLBAR_ICONS:
    case GTK_TOOLBAR_TEXT:
    case GTK_TOOLBAR_BOTH:
	break;
    case GTK_TOOLBAR_BOTH_HORIZ:
	i = GTK_TOOLBAR_BOTH;
	break;
    default:  /* illegal -- ignore */
	gtkpod_warning (_("prefs_set_toolbar_style: illegal style '%d' ignored\n"), i);
	return;
    }

    cfg->toolbar_style = i;
    display_show_hide_toolbar ();
}

gboolean prefs_get_pm_autostore (void)
{
    return cfg->sortcfg.pm_autostore;
}

void prefs_set_pm_autostore (gboolean val)
{
    cfg->sortcfg.pm_autostore = val;
}

gboolean prefs_get_tm_autostore (void)
{
    return cfg->sortcfg.tm_autostore;
}

void prefs_set_tm_autostore (gboolean val)
{
    cfg->sortcfg.tm_autostore = val;
}

gint prefs_get_pm_sort (void)
{
    return cfg->sortcfg.pm_sort;
}

void prefs_set_pm_sort (gint i)
{
    switch (i)
    {
    case SORT_ASCENDING:
    case SORT_DESCENDING:
    case SORT_NONE:
	break;
    default:  /* illegal -- ignore */
	gtkpod_warning (_("prefs_set_pm_sort: illegal type '%d' ignored\n"), i);
	return;
    }

    cfg->sortcfg.pm_sort = i;
}

gint prefs_get_st_sort (void)
{
    return cfg->sortcfg.st_sort;
}

void prefs_set_st_sort (gint i)
{
    switch (i)
    {
    case SORT_ASCENDING:
    case SORT_DESCENDING:
    case SORT_NONE:
	cfg->sortcfg.st_sort = i;
	break;
    default:  /* illegal -- ignore */
	gtkpod_warning ("Programming error: prefs_set_st_sort: illegal type '%d' ignored\n", i);
	break;
    }
}

gint prefs_get_tm_sort (void)
{
    return cfg->sortcfg.tm_sort;
}

void prefs_set_tm_sort (gint i)
{
    switch (i)
    {
    case SORT_ASCENDING:
    case SORT_DESCENDING:
    case SORT_NONE:
	cfg->sortcfg.tm_sort = i;
	break;
    default:  /* illegal -- ignore */
	gtkpod_warning ("Programming error: prefs_set_tm_sort: illegal type '%d' ignored\n", i);
	break;
    }

}

TM_item prefs_get_tm_sortcol (void)
{
    return cfg->sortcfg.tm_sortcol;
}

void prefs_set_tm_sortcol (TM_item i)
{
    if (i < TM_NUM_COLUMNS)
	cfg->sortcfg.tm_sortcol = i;
}

void prefs_set_display_tooltips_main (gboolean state)
{
    cfg->display_tooltips_main = state;
    display_show_hide_tooltips ();
}

gboolean prefs_get_display_tooltips_main (void)
{
    return cfg->display_tooltips_main;
}

void prefs_set_display_tooltips_prefs (gboolean state)
{
    cfg->display_tooltips_prefs = state;
    display_show_hide_tooltips ();
}

gboolean prefs_get_display_tooltips_prefs (void)
{
    return cfg->display_tooltips_prefs;
}

void prefs_set_multi_edit (gboolean state)
{
    cfg->multi_edit = state;
}

gboolean prefs_get_multi_edit (void)
{
    return cfg->multi_edit;
}

void prefs_set_not_played_track (gboolean state)
{
    cfg->not_played_track = state;
}

gboolean prefs_get_not_played_track (void)
{
    return cfg->not_played_track;
}

void prefs_set_misc_track_nr (gint state)
{
    cfg->misc_track_nr = state;
}

gint prefs_get_misc_track_nr (void)
{
    return cfg->misc_track_nr;
}

void prefs_set_multi_edit_title (gboolean state)
{
    cfg->multi_edit_title = state;
}

gboolean prefs_get_multi_edit_title (void)
{
    return cfg->multi_edit_title;
}

gint prefs_get_last_prefs_page (void)
{
    return cfg->last_prefs_page;
}

void prefs_set_last_prefs_page (gint i)
{
    cfg->last_prefs_page = i;
}

/* validate the the play_path @path and return a valid copy that has
 * to be freed with g_free when it's not needed any more. */
/* Rules: - must be '%(member of allowed)'
	  - removes all invalid '%' */
gchar *prefs_validate_path (const gchar *path, const gchar *allowed)
{
    const gchar *pp;
    gchar *npp, *npath=NULL;

    if ((!path) || (strlen (path) == 0)) return g_strdup ("");
    if (!allowed)  allowed = "";

    npath = g_malloc0 (strlen (path)+1); /* new path can only be shorter
					    than old path */
    pp = path;
    npp = npath;
    while (*pp)
    {
	if (*pp == '%')
	{
	    if (*(pp+1) && strchr (allowed, *(pp+1)) == NULL)
	    {
		if (strlen (allowed) == 0)
		{
		    gtkpod_warning (_("'%s': no arguments (%%...) allowed.\n"),
				    path);
		}
		else
		{
		    gtkpod_warning (_("'%s': only '%%[%s]' allowed.\n"),
				    path, allowed);
		}
		++pp; /* skip '%...' */
	    }
	    else
	    {   /* copy '%s' */
		*npp++ = *pp++;
		*npp++ = *pp;
	    }
	}
	else
	{
	    *npp++ = *pp;
	}
	if (*pp) ++pp; /* increment if we are not at the end */
    }
    return npath;
}


void prefs_set_path (PathType i, const gchar *path)
{
    switch (i)
    {
    case PATH_PLAY_NOW:
    case PATH_PLAY_ENQUEUE:
	g_free (cfg->path[i]);
	cfg->path[i] = prefs_validate_path (path, "s");
	break;
    case PATH_MP3GAIN:
    case PATH_MSERV_MUSIC_ROOT:
    case PATH_MSERV_TRACKINFO_ROOT:
	g_free (cfg->path[i]);
	if (path)
	    cfg->path[i] = g_strstrip (g_strdup (path));
	else
	    cfg->path[i] = g_strdup ("");
	break;
    case PATH_SYNC_CONTACTS:
    case PATH_SYNC_CALENDAR:
    case PATH_SYNC_NOTES:
	g_free (cfg->path[i]);
	cfg->path[i] = prefs_validate_path (path, "i");
	break;
    case PATH_NUM:
	break;
    }
}


const gchar *prefs_get_path (PathType i)
{
    g_return_val_if_fail (i>=0 && i<PATH_NUM, "");
    return cfg->path[i];
}

gboolean 
prefs_get_automount (void)
{
    return cfg->automount;
}
void

prefs_set_automount(gboolean val)
{
    cfg->automount = val;
}

gboolean
prefs_get_info_window(void)
{
    return cfg->info_window;
}
void

prefs_set_info_window(gboolean val)
{
    cfg->info_window = val;
}

void prefs_set_sp_or (guint32 inst, gboolean state)
{
    if (inst < SORT_TAB_MAX)	cfg->st[inst].sp_or = state;
}

gboolean prefs_get_sp_or (guint32 inst)
{
    if (inst < SORT_TAB_MAX)	return cfg->st[inst].sp_or;
    return FALSE;
}

/* Set whether condition @t_item in sort tab @inst is activated or not */
void prefs_set_sp_cond (guint32 inst, T_item t_item, gboolean state)
{
    if (inst < SORT_TAB_MAX)
    {
	switch (t_item)
	{
	case T_RATING:
	    cfg->st[inst].sp_rating = state;
	    break;
	case T_PLAYCOUNT:
	    cfg->st[inst].sp_playcount = state;
	    break;
	case T_TIME_PLAYED:
	    cfg->st[inst].sp_played = state;
	    break;
	case T_TIME_MODIFIED:
	    cfg->st[inst].sp_modified = state;
	    break;
	case T_TIME_ADDED:
	    cfg->st[inst].sp_added = state;
	    break;
	default:
	    /* programming error */
	    fprintf (stderr, "prefs_set_sp_cond(): inst=%d, !t_item=%d!\n",
		     inst, t_item);
	    break;
	}
    }
    else
    {
	/* programming error */
	fprintf (stderr, "prefs_set_sp_cond(): !inst=%d! t_item=%d\n",
		 inst, t_item);
    }
}


/* Set whether condition @t_item in sort tab @inst is activated or not */
gboolean prefs_get_sp_cond (guint32 inst, T_item t_item)
{
    if (inst < SORT_TAB_MAX)
    {
	switch (t_item)
	{
	case T_RATING:
	    return cfg->st[inst].sp_rating;
	case T_PLAYCOUNT:
	    return cfg->st[inst].sp_playcount;
	case T_TIME_PLAYED:
	    return cfg->st[inst].sp_played;
	case T_TIME_MODIFIED:
	    return cfg->st[inst].sp_modified;
	case T_TIME_ADDED:
	    return cfg->st[inst].sp_added;
	default:
	    /* programming error */
	    fprintf (stderr, "prefs_get_sp_cond(): inst=%d !t_item=%d!\n",
		     inst, t_item);
	    break;
	}
    }
    else
    {
	/* programming error */
	fprintf (stderr, "prefs_get_sp_cond(): !inst=%d! t_item=%d\n",
		 inst, t_item);
    }
    return FALSE;
}


void prefs_set_sp_rating_n (guint32 inst, gint n, gboolean state)
{
    if ((inst < SORT_TAB_MAX) && (n <=RATING_MAX))
    {
	if (state)
	    cfg->st[inst].sp_rating_state |= (1<<n);
	else
	    cfg->st[inst].sp_rating_state &= ~(1<<n);
    }
    else
	fprintf (stderr, "prefs_set_sp_rating_n(): inst=%d, n=%d\n", inst, n);
}


gboolean prefs_get_sp_rating_n (guint32 inst, gint n)
{
    if ((inst < SORT_TAB_MAX) && (n <=RATING_MAX))
    {
	if ((cfg->st[inst].sp_rating_state & (1<<n)) != 0)
	    return TRUE;
	else
	    return FALSE;
    }
    fprintf (stderr, "prefs_get_sp_rating_n(): inst=%d, n=%d\n", inst, n);
    return FALSE;
}


void prefs_set_sp_rating_state (guint32 inst, guint32 state)
{
    if (inst < SORT_TAB_MAX)
	/* only keep the 'RATING_MAX+1' lowest bits */
	cfg->st[inst].sp_rating_state = (state & ((1<<(RATING_MAX+1))-1));
    else
	fprintf (stderr, "prefs_set_sp_rating_state(): inst=%d\n", inst);
}


guint32 prefs_get_sp_rating_state (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
	return cfg->st[inst].sp_rating_state;
    fprintf (stderr, "prefs_get_sp_rating_state(): inst=%d\n", inst);
    return 0;
}


void prefs_set_sp_entry (guint32 inst, T_item t_item, const gchar *str)
{
/*    printf("psse: %d, %d, %s\n", inst, t_item, str);*/
    if (inst < SORT_TAB_MAX)
    {
	gchar *cstr = NULL;

	if (str)  cstr = g_strdup (str);

	switch (t_item)
	{
	case T_TIME_PLAYED:
	    g_free (cfg->st[inst].sp_played_state);
	    cfg->st[inst].sp_played_state = cstr;
	    break;
	case T_TIME_MODIFIED:
	    g_free (cfg->st[inst].sp_modified_state);
	    cfg->st[inst].sp_modified_state = cstr;
	    break;
	case T_TIME_ADDED:
	    g_free (cfg->st[inst].sp_added_state);
	    cfg->st[inst].sp_added_state = cstr;
	    break;
	default:
	    /* programming error */
	    g_free (cstr);
	    fprintf (stderr, "prefs_set_sp_entry(): inst=%d !t_item=%d!\n",
		     inst, t_item);
	    break;
	}
    }
    else
    {
	/* programming error */
	fprintf (stderr, "prefs_set_sp_entry(): !inst=%d! t_item=%d\n",
		 inst, t_item);
    }
}


/* Returns the current default or a pointer to "". Guaranteed to never
   return NULL */
gchar *prefs_get_sp_entry (guint32 inst, T_item t_item)
{
    gchar *result = NULL;

    if (inst < SORT_TAB_MAX)
    {
	switch (t_item)
	{
	case T_TIME_PLAYED:
	    result = cfg->st[inst].sp_played_state;
	    break;
	case T_TIME_MODIFIED:
	    result = cfg->st[inst].sp_modified_state;
	    break;
	case T_TIME_ADDED:
	    result = cfg->st[inst].sp_added_state;
	    break;
	default:
	    /* programming error */
	    fprintf (stderr, "prefs_get_sp_entry(): !t_item=%d!\n", t_item);
	    break;
	}
    }
    else
    {
	/* programming error */
	fprintf (stderr, "prefs_get_sp_entry(): !inst=%d!\n", inst);
    }
    if (result == NULL)    result = "";
    return result;
}


void prefs_set_sp_autodisplay (guint32 inst, gboolean state)
{
    if (inst < SORT_TAB_MAX)
	cfg->st[inst].sp_autodisplay = state;
    else
	fprintf (stderr, "prefs_set_sp_autodisplay(): !inst=%d!\n", inst);
}


gboolean prefs_get_sp_autodisplay (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
	return cfg->st[inst].sp_autodisplay;
    else
	fprintf (stderr, "prefs_get_sp_autodisplay(): !inst=%d!\n", inst);
    return FALSE;
}


gint32 prefs_get_sp_playcount_low (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
	return cfg->st[inst].sp_playcount_low;
    else
	fprintf (stderr, "prefs_get_sp_playcount_low(): !inst=%d!\n", inst);
    return 0;
}


gint32 prefs_get_sp_playcount_high (guint32 inst)
{
    if (inst < SORT_TAB_MAX)
	return cfg->st[inst].sp_playcount_high;
    else
	fprintf (stderr, "prefs_get_sp_playcount_high(): !inst=%d!\n", inst);
    return (guint32)-1;
}


void prefs_set_sp_playcount_low (guint32 inst, gint32 limit)
{
    if (inst < SORT_TAB_MAX)
	cfg->st[inst].sp_playcount_low = limit;
    else
	fprintf (stderr, "prefs_set_sp_playcount_low(): !inst=%d!\n", inst);
}


void prefs_set_sp_playcount_high (guint32 inst, gint32 limit)
{
    if (inst < SORT_TAB_MAX)
	cfg->st[inst].sp_playcount_high = limit;
    else
	fprintf (stderr, "prefs_set_sp_playcount_high(): !inst=%d!\n", inst);
}

gboolean prefs_get_concal_autosync(void)
{
    return(cfg->concal_autosync);
}

void prefs_set_concal_autosync(gboolean val)
{
    cfg->concal_autosync = val;
}

gboolean prefs_get_tmp_disable_sort(void)
{
    return(cfg->tmp_disable_sort);
}

void prefs_set_tmp_disable_sort(gboolean val)
{
    cfg->tmp_disable_sort = val;
}

gboolean prefs_get_startup_messages(void)
{
    return(cfg->startup_messages);
}

void prefs_set_startup_messages(gboolean val)
{
    cfg->startup_messages = val;
}

/* sorting gets disabled temporarily if either of the options
   'tmp_disable_sort' or 'block_display' is checked */
gboolean prefs_get_disable_sorting(void)
{
    return (prefs_get_block_display() || prefs_get_tmp_disable_sort());
}

/* whether or not to read ratings from mserv db */
gboolean prefs_get_mserv_use(void)
{
    return(cfg->mserv_use);
}

void prefs_set_mserv_use(gboolean val)
{
    cfg->mserv_use = val;
}

/* whether or not to report problems reading from mserv database */
gboolean prefs_get_mserv_report_probs(void)
{
    return(cfg->mserv_report_probs);
}

void prefs_set_mserv_report_probs(gboolean val)
{
    cfg->mserv_report_probs = val;
}

void prefs_set_mserv_username (const gchar *str)
{
    if (str)
    {
	g_free (cfg->mserv_username);
	cfg->mserv_username = g_strdup (str);
    }
}

const gchar *prefs_get_mserv_username (void)
{
    return cfg->mserv_username;
}

gboolean prefs_get_unused_gboolean3(void)
{
    return(cfg->unused_gboolean3);
}

void prefs_set_unused_gboolean3(gboolean val)
{
    cfg->unused_gboolean3 = val;
}

void prefs_set_pc_dir (const gchar *str)
{
    g_return_if_fail (str);

    g_free (cfg->pc_dir);
    cfg->pc_dir = convert_filename (str);
}

/* g_free() after use */
gchar *prefs_get_pc_dir (void)
{
    return g_strdup (cfg->pc_dir);
}

void prefs_set_pc_del_age(gboolean val)
{
    cfg->pc_del_age = val;
}

gboolean prefs_get_pc_del_age(void)
{
    return cfg->pc_del_age;
}

void prefs_set_pc_del_age_val(gint val)
{
    cfg->pc_del_age_val = val;
}

gint prefs_get_pc_del_age_val(void)
{
    return cfg->pc_del_age_val;
}

void prefs_set_pc_del_copied(gboolean val)
{
    cfg->pc_del_copied = val;
}

gboolean prefs_get_pc_del_copied(void)
{
    return cfg->pc_del_copied;
}

void prefs_set_pc_auto_fetch(gboolean val)
{
    cfg->pc_auto_fetch = val;
}

gboolean prefs_get_pc_auto_fetch(void)
{
    return cfg->pc_auto_fetch;
}

void prefs_set_pc_log(gboolean val)
{
    cfg->pc_log = val;
}

gboolean prefs_get_pc_log(void)
{
    return cfg->pc_log;
}

void prefs_set_pc_log_file(const gchar *str)
{
    if (str)
    {
        g_free (cfg->pc_log_file);
        cfg->pc_log_file = convert_filename (str);
    }
}

/* g_free() after use */
gchar *prefs_get_pc_log_file(void)
{
    return g_strdup (cfg->pc_log_file);
}

void prefs_set_pc_auto_sync(gboolean val)
{
    cfg->pc_auto_sync = val;
}

gboolean prefs_get_pc_auto_sync(void)
{
    return cfg->pc_auto_sync;
}

void prefs_set_pc_ipod_del_age(gboolean val)
{
    cfg->pc_ipod_del_age = val;
}

gboolean prefs_get_pc_ipod_del_age(void)
{
    return cfg->pc_ipod_del_age;
}

void prefs_set_pc_ipod_del_age_val(gint val)
{
    cfg->pc_ipod_del_age_val = val;
}

gint prefs_get_pc_ipod_del_age_val(void)
{
    return cfg->pc_ipod_del_age_val;
}

void prefs_set_pc_ipod_del_played(gboolean val)
{
    cfg->pc_ipod_del_played = val;
}

gboolean prefs_get_pc_ipod_del_played(void)
{
    return cfg->pc_ipod_del_played;
}

void prefs_set_pc_ipod_inc_date(gboolean val)
{
    cfg->pc_ipod_inc_date = val;
}

gboolean prefs_get_pc_ipod_inc_date(void)
{
    return cfg->pc_ipod_inc_date;
}

void prefs_set_pc_change_genre(gboolean val)
{
    cfg->pc_change_genre = val;
}

gboolean prefs_get_pc_change_genre(void)
{
    return cfg->pc_change_genre;
}

/* ------------------------------------------------------------

   Functions for generic preferences settings

   ------------------------------------------------------------ */

GHashTable *prefs_hash = NULL;

/* Set a string setting (e.g. @key="export_path") to @value */
void prefs_set_string_value (const gchar *key, const gchar *value)
{
    if (!key)  return;

    if (!prefs_hash)
	prefs_hash = g_hash_table_new_full (g_str_hash, g_str_equal,
					    g_free, g_free);

    if (!value)
    {
	g_hash_table_remove (prefs_hash, key);
    }
    else
    {
	g_hash_table_insert (prefs_hash,
			     g_strdup (key), g_strdup (value));
    }
}


/* Set a integer setting (e.g. @key="file_type") to @value */
void prefs_set_int_value (const gchar *key, gint value)
{
    gchar *str;

    if (!key)  return;

    str = g_strdup_printf ("%d", value);
    prefs_set_string_value (key, str);
    g_free (str);
}


/* Set a integer setting (e.g. @key="file_type") to @value */
void prefs_set_int64_value (const gchar *key, gint64 value)
{
    gchar *str;

    if (!key)  return;

    str = g_strdup_printf ("%llu", value);
    prefs_set_string_value (key, str);
    g_free (str);
}


/* Retrieve a string setting. @value will be filled with a copy of the
 * value. */
/* Return value: TRUE, if @key could be retrieved. Otherwise FALSE and
 * @value is set to NULL */
gboolean prefs_get_string_value (const gchar *key, gchar **value)
{
    g_return_val_if_fail (value, FALSE);

    *value = NULL;
    if (prefs_hash && key)
    {
	*value = g_strdup (g_hash_table_lookup (prefs_hash, key));
    }
    return (*value != NULL);
}


/* Same as prefs_get_string_value() but easier to use. The string is
   simply returned, a returned NULL value means that the @key could
   not be found. g_free() the result when it's no longer needed */
gchar *prefs_get_string (const gchar *key)
{
    if (prefs_hash && key)
	return g_strdup (g_hash_table_lookup (prefs_hash, key));
    return NULL;
}


/* Retrieve an integer setting. @value will be filled with the stored
 * integer value. */
/* Return value: TRUE, if @key could be retrieved. Otherwise FALSE and
 * @value is set to 0 */
gboolean prefs_get_int_value (const gchar *key, gint *value)
{
    g_return_val_if_fail (value, FALSE);

    *value = 0;
    if (prefs_hash && key)
    {
	gchar *str = g_hash_table_lookup (prefs_hash, key);
	if (str)
	{
	    *value = atoi (str);
	    return TRUE;
	}
    }
    return FALSE;
}


/* Retrieve an integer setting. */
/* Return value: the set integer value or 0 if no value is set. */
/* Use prefs_get_int_value if you need to know whether a value was
   actually set in the prefs or not. */
gint prefs_get_int (const gchar *key)
{
    if (prefs_hash && key)
    {
	gchar *str = g_hash_table_lookup (prefs_hash, key);
	if (str)
	    return atoi (str);
    }
    return 0;
}


/* Retrieve a 64 bit integer setting. @value will be filled with the
 * stored integer value. */
/* Return value: TRUE, if @key could be retrieved. Otherwise FALSE and
 * @value is set to 0 */
gboolean prefs_get_int64_value (const gchar *key, gint64 *value)
{
    g_return_val_if_fail (value, FALSE);

    *value = 0;
    if (prefs_hash && key)
    {
	gchar *str = g_hash_table_lookup (prefs_hash, key);
	if (str)
	{
	    *value = (gint64)g_ascii_strtoull (str, NULL, 10);
	    return TRUE;
	}
    }
    return FALSE;
}


/* Retrieve a 64 bit integer setting. */
/* Return value: the set integer value or 0 if no value is set. */
/* Use prefs_get_int64_value if you need to know whether a value was
   actually set in the prefs or not. */
gint64 prefs_get_int64 (const gchar *key)
{
    if (prefs_hash && key)
    {
	gchar *str = g_hash_table_lookup (prefs_hash, key);
	if (str)
	    return (gint64)g_ascii_strtoull (str, NULL, 10);
    }
    return 0;
}


/* write the values in the hash table out into the preferences file
   @fp */
static void prefs_write_hash_values (FILE *fp)
{
    void prefs_write_hash_func (gpointer key, gpointer value,
				gpointer fp)
	{
	    g_return_if_fail (key && value && fp);
	    fprintf ((FILE *)fp, "%s=%s\n", (gchar *)key, (gchar *)value);
	}

    if (prefs_hash)
	g_hash_table_foreach (prefs_hash, prefs_write_hash_func, fp);
}
